
static char help[] = "Tests DMSwarm with DMShell\n\n";

#include <petscsf.h>
#include <petscdm.h>
#include <petscdmshell.h>
#include <petscdmda.h>
#include <petscdmswarm.h>

typedef struct{
    PetscInt xs,ys,zs,
     	     xm,ym,zm;
} local_corners_struct;

typedef struct {
    PetscReal T;        // temperature at time n+1 (8 bytes)
    PetscInt  B;        //(4 bytes)
    //vel_struct v_new;    // velocity at time n+1
    //vel_struct v_old;    // velocity at time n
    //vel_struct Resv;     // velocity residuals 
    //PetscReal ResT;     // temperature residuals
} Particle_struct;

/*
 Create a DMShell and attach a regularly spaced DMDA for point location
 Override methods for point location
*/
PetscErrorCode ex3_1(void)
{
    PetscErrorCode        ierr;
    PetscInt              n,m,l,part_nbr;
    PetscInt              bs, bp;
    DM                    da, dmcell, swarm; 
    PetscReal             xmin, xmax, ymin, ymax, zmin, zmax, *coords, *partInfo;
    PetscInt              *cell_id, *rankval, *particle_id;
    PetscMPIInt           rank,size;
    local_corners_struct  corners;
    PetscInt              npoints[2],npoints_orig[2];
    PetscRandom           rand;
    Particle_struct       *partStruct;
    
    xmin = 0.0; ymin = 0.0; zmin = 0.0; xmax = 1.0; ymax = 1.0; zmax = 1.0;
    n = 4; m = 4; l = 4; 
    part_nbr = 10;                                                             // Total number of particles
    
    /* Create a regularly spaced DMDA */
    ierr = DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_NONE, DM_BOUNDARY_NONE, DM_BOUNDARY_NONE,DMDA_STENCIL_BOX,n,m,l,PETSC_DECIDE,1,PETSC_DECIDE,1,1,0,0,0,&da);CHKERRQ(ierr);    
    //ierr = DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,DMDA_STENCIL_BOX,n,m,l,PETSC_DECIDE,1,PETSC_DECIDE,1,1,0,0,0,&da);CHKERRQ(ierr);
    ierr = DMSetFromOptions(da);CHKERRQ(ierr);
    ierr = DMSetUp(da);CHKERRQ(ierr);
    DMDASetUniformCoordinates(da,xmin, xmax, ymin, ymax, zmin, zmax);

    /* Processor Info */
    MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
    MPI_Comm_size(PETSC_COMM_WORLD,&size);
    PetscPrintf(PETSC_COMM_WORLD,"Number of processors = %d, rank = %d\n",size,rank);
    DMDAGetCorners(da,&corners.xs,&corners.ys,&corners.zs,&corners.xm,&corners.ym,&corners.zm);
    PetscPrintf(PETSC_COMM_SELF,"[Proc %d] cornx = %d, corny = %d, cornz = %d, cornxm = %d, cornym = %d, cornzm = %d\n",rank,corners.xs,corners.ys,corners.zs,corners.xm,corners.ym,corners.zm);
	
    /* Cells coordinates */
    DMDACoor3d ***coors;
    Vec      vcoors;
    DM       cda;
    PetscReal x,y,z;
    DMGetCoordinates(da,&vcoors);
    DMGetCoordinateDM(da,&cda);
    DMDAVecGetArray(cda,vcoors,&coors);
    // X Coordinates:
    PetscPrintf(PETSC_COMM_SELF,"[Proc %d] X = ",rank);
    for (int i=corners.xs; i<corners.xs+corners.xm; i++) {
        x = coors[corners.zs][corners.ys][i].x;
        PetscPrintf(PETSC_COMM_SELF,"%f ",x);
    }
    PetscPrintf(PETSC_COMM_SELF,"\n");
    // Y Coordinates:
    PetscPrintf(PETSC_COMM_SELF,"[Proc %d] Y = ",rank);
    for (int j=corners.ys; j<corners.ys+corners.ym; j++) {
        y = coors[corners.zs][j][corners.xs].y;
        PetscPrintf(PETSC_COMM_SELF,"%f ",y);
    }
    PetscPrintf(PETSC_COMM_SELF,"\n");
    // Z Coordinates:
    PetscPrintf(PETSC_COMM_SELF,"[Proc %d] Z = ",rank);
    for (int k=corners.zs; k<corners.zs+corners.zm; k++) {
        z = coors[k][corners.ys][corners.xs].z;
        PetscPrintf(PETSC_COMM_SELF,"%f ",z);
    }
    PetscPrintf(PETSC_COMM_SELF,"\n");
    DMDAVecRestoreArray(cda,vcoors,&coors);

    /* Create a DMShell for point location purposes */ 
    ierr = DMShellCreate(PETSC_COMM_WORLD,&dmcell);CHKERRQ(ierr);
    ierr = DMSetApplicationContext(dmcell,da);CHKERRQ(ierr);
    //dmcell->ops->locatepoints = DMLocatePoints_DMDARegular;
    //dmcell->ops->getneighbors = DMGetNeighbors_DMDARegular;
    PetscPrintf(PETSC_COMM_WORLD,"HERE");

    /* Create the swarm */
    DMCreate(PETSC_COMM_WORLD,&swarm);
    DMSetType(swarm,DMSWARM);
    DMSetDimension(swarm,3);

    DMSwarmSetType(swarm,DMSWARM_PIC);
    DMSwarmSetCellDM(swarm,da);

    DMSwarmRegisterPetscDatatypeField(swarm,"particle_field",1,PETSC_REAL);
	  DMSwarmRegisterUserStructField(swarm,"particle_struct",16); // Will store Particle_struct for each particle
    DMSwarmFinalizeFieldRegister(swarm);

    /* Assign all particles to rank 0 */
	  PetscPrintf(PETSC_COMM_WORLD,"All %d particles are stored on rank 0 and assigned random coordinates inside the whole domain \n", part_nbr);
    PetscInt part_proc = (int) part_nbr * corners.xm * corners.zm / n / l;
    if (rank == 0){
        DMSwarmSetLocalSizes(swarm,part_nbr,part_nbr);
    }else{
        DMSwarmSetLocalSizes(swarm,0,part_nbr);
    }

    //Particle_struct       *particle_info;
    /* Assign random (x,y,z) coordinates (located inside the whole domain) */
    ierr = DMSwarmGetLocalSize(swarm,&npoints_orig[0]);CHKERRQ(ierr);
    ierr = DMSwarmGetSize(swarm,&npoints_orig[1]);CHKERRQ(ierr);
    ierr = DMSwarmGetField(swarm,DMSwarmPICField_coor,&bs,NULL,(void**)&coords);CHKERRQ(ierr);
    ierr = PetscRandomCreate(PETSC_COMM_SELF,&rand);CHKERRQ(ierr);
    ierr = PetscRandomSetInterval(rand,0.0,1.0);CHKERRQ(ierr);
    PetscInt cnt = 0; // MUST TRY TO GET RID OF IT
    for (int p=0; p<npoints_orig[0]; p++) {
      PetscReal rx,ry,rz;

      ierr = PetscRandomGetValueReal(rand,&rx);CHKERRQ(ierr);
      ierr = PetscRandomGetValueReal(rand,&ry);CHKERRQ(ierr);
      ierr = PetscRandomGetValueReal(rand,&rz);CHKERRQ(ierr);

      coords[bs*cnt] = rx * (xmax-xmin) + xmin;
      coords[bs*cnt+1] = ry * (ymax-ymin) + ymin;
      coords[bs*cnt+2] = rz * (zmax-zmin) + zmin;
      cnt++;
    }
    ierr = PetscRandomDestroy(&rand);CHKERRQ(ierr);
    ierr = DMSwarmRestoreField(swarm,DMSwarmPICField_coor,&bs,NULL,(void**)&coords);CHKERRQ(ierr);
    
    ierr = DMSwarmGetField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);CHKERRQ(ierr);
    ierr = DMSwarmGetField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);CHKERRQ(ierr);
    for (int p=0; p<npoints_orig[0]; p++) {
      partInfo[p] = 55.4 + p;
      partStruct[p].T = 1.5 + p;
      partStruct[p].B = 3 + p;
    }
    ierr = DMSwarmRestoreField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);CHKERRQ(ierr);
    ierr = DMSwarmRestoreField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);CHKERRQ(ierr);

    /* Print each particles coordinates, T, R */
    DMSwarmGetField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmGetField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);
    DMSwarmGetField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);
    //DMSwarmGetField(swarm,"particle_field",&bp,NULL,(void**)&particle_info);
    for (int i=0; i<npoints_orig[0] ; i++){
        PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f \n", rank, coords[i*3], coords[i*3+1], coords[i*3+2]);
        //PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f, part = %f , T = %f, B = %d \n", rank, coords[i*3], coords[i*3+1], coords[i*3+2], partInfo[i], partStruct[i].T, partStruct[i].B);
    }
    DMSwarmRestoreField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmRestoreField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);CHKERRQ(ierr);
    DMSwarmRestoreField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);
    //DMSwarmRestoreField(swarm,"particle_field",&bp,NULL,(void**)&particle_info);

    /* Migrate points */
    ierr = DMSwarmMigrate(swarm,PETSC_TRUE);CHKERRQ(ierr);

    DMSwarmGetLocalSize(swarm,&npoints[0]);
    DMSwarmGetSize(swarm,&npoints[1]);

    /* Print the before after number of particles */
    PetscViewerASCIIPushSynchronized(PETSC_VIEWER_STDOUT_WORLD);
    PetscViewerASCIISynchronizedPrintf(PETSC_VIEWER_STDOUT_WORLD,"rank[%d] before(%D out of %D) after(%D out of %D)\n",rank,npoints_orig[0],npoints_orig[1],npoints[0],npoints[1]);
    PetscViewerFlush(PETSC_VIEWER_STDOUT_WORLD);
    PetscViewerASCIIPopSynchronized(PETSC_VIEWER_STDOUT_WORLD);

    /* Print each particles coordinates, T, R */
    DMSwarmGetField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmGetField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);
    DMSwarmGetField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);
    DMSwarmGetField(swarm,"DMSwarm_rank",NULL,NULL,(void**)&rankval);
    DMSwarmGetField(swarm,"DMSwarm_pid",NULL,NULL,(void**)&particle_id);
    DMSwarmGetField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);
    //DMSwarmGetField(swarm,"particle_field",&bp,NULL,(void**)&particle_info);
    for (int i=0; i<npoints[0] ; i++){
		PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f, \t cell_id =  %d,\t rank = %d,\t particle_id = %D\n", rank, coords[i*3], coords[i*3+1], coords[i*3+2], cell_id[i], rankval[i], particle_id[i]); //,particle_info[i].T);
        //PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f, part = %f, T=%f, B = %d, \t cell_id =  %d,\t rank = %d,\t particle_id = %D\n", rank, coords[i*3], coords[i*3+1], coords[i*3+2], partInfo[i], partStruct[i].T, partStruct[i].B, cell_id[i], rankval[i], particle_id[i]); //,particle_info[i].T);
    
    }
    DMSwarmRestoreField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmRestoreField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);
    DMSwarmRestoreField(swarm,"particle_field",NULL,NULL,(void**)&partInfo);
    DMSwarmRestoreField(swarm,"DMSwarm_rank",NULL,NULL,(void**)&rankval);
    DMSwarmRestoreField(swarm,"DMSwarm_pid",NULL,NULL,(void**)&particle_id);
    DMSwarmRestoreField(swarm,"particle_struct",NULL,NULL,(void**)&partStruct);
    //DMSwarmRestoreField(swarm,"particle_field",&bp,NULL,(void**)&particle_info);


    // PRINT TO FILE
    Vec coordinates;
    DMSwarmCreateGlobalVectorFromField(swarm,"DMSwarmPIC_coor", &coordinates);
    //DMSwarmCreateGlobalVectorFromField(swarm,"particle_struct", partStruct); // WILL NOT WORK WITH STRUCTURE!! NEED VEC
    char coord_path[128];
    PetscViewer coord_view;
    char temp_path[128];
    PetscViewer temp_view; 

    sprintf(coord_path,"results/%d_coord.txt",0);
    PetscViewerBinaryOpen(PETSC_COMM_WORLD,coord_path,FILE_MODE_WRITE,&coord_view);
	VecView(coordinates,coord_view);

    sprintf(temp_path,"results/%d_temp.txt",0);
    //PetscViewerBinaryOpen(PETSC_COMM_WORLD,temp_path,FILE_MODE_WRITE,&temp_view);
	//VecView(partStruct.T,temp_view);

    DMSwarmDestroyGlobalVectorFromField(swarm,"DMSwarmPIC_coor",&coordinates);
    //DMSwarmDestroyGlobalVectorFromField(swarm,"particle_struct",partStruct);

    PetscViewerDestroy(&coord_view);
    //PetscViewerDestroy(&temp_view);

    ierr = DMDestroy(&da);CHKERRQ(ierr);
    ierr = DMDestroy(&dmcell);CHKERRQ(ierr);
    ierr = DMDestroy(&swarm);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}

int main(int argc,char **argv)
{
  PetscErrorCode ierr;

  ierr = PetscInitialize(&argc,&argv,(char*)0,help);if (ierr) return ierr;
  ierr = ex3_1();CHKERRQ(ierr);
  ierr = PetscFinalize();
  return ierr;
}

/*TEST

   build:
      requires: double !complex

   test:
      filter: grep -v atomic
      filter_output: grep -v atomic
TEST*/
