
static char help[] = "Tests DMSwarm with DMShell\n\n";

#include <petscsf.h>
#include <petscdm.h>
#include <petscdmshell.h>
#include <petscdmda.h>
#include <petscdmswarm.h>
#include <petscksp.h>

typedef struct{
    PetscInt xs,ys,zs,
     	     xm,ym,zm;
} local_corners_struct;

PetscErrorCode ex3_3(void){

    PetscInt M_coll,N_coll,P_coll;
    M_coll = N_coll = P_coll = 10;
    DM da, da_swarm;
    int part_nbr = 10;
    MPI_Comm         comm;
    PetscInt         istart, jstart, kstart, isize, jsize, ksize; 
    DM               cda, swarm;
    Vec              xcoor;
    PetscScalar      *coors;
    PetscInt cnt; 

    PetscInt rank,size;
	MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	MPI_Comm_size(PETSC_COMM_WORLD,&size);
    
    // Create a collocated DMDA with ghost cells
    DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,DMDA_STENCIL_BOX,M_coll,N_coll,P_coll,PETSC_DECIDE,1,PETSC_DECIDE,1,1,0,0,0,&da);
	DMSetFromOptions(da);
    DMSetUp(da);

    // Get da info
    PetscInt dim, M,N,P,m,n,p,dof,s;
    DMBoundaryType bx,by,bz;
    DMDAStencilType st;
    DMDAGetInfo(da,&dim,&M,&N,&P,&m,&n,&p,&dof,&s,&bx,&by,&bz,&st);

    // Add 1 cell in each direction
    M = M + 1;
    N = N + 1;
    P = P + 1;

    // Get ownership ranges
    const PetscInt *lx_init, *ly_init, *lz_init;
    DMDAGetOwnershipRanges(da,&lx_init,&ly_init,&lz_init);

    // Add 1 cell to the ownership ranges
    PetscInt *lx = new PetscInt[n];
    PetscInt *ly = new PetscInt[m];
    PetscInt *lz = new PetscInt[p];
    for(int i = 0; i < m; i++){
        lx[i] = lx_init[i];
    }
    for(int i = 0; i < n; i++){
        ly[i] = ly_init[i];
    }
    for(int i = 0; i < p; i++){
        lz[i] = lz_init[i];
    }
    lx[0] = lx_init[0] + 1;
    ly[0] = ly_init[0] + 1;
    lz[0] = lz_init[0] + 1; 

    // Create a DMDA staggered and without ghost cells (for DMSwarm to work)
    //DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_NONE, DM_BOUNDARY_NONE, DM_BOUNDARY_NONE,DMDA_STENCIL_BOX,M,N,P,PETSC_DECIDE,1,PETSC_DECIDE,1,1,0,0,0,&da_swarm);
    DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_NONE, DM_BOUNDARY_NONE, DM_BOUNDARY_NONE,DMDA_STENCIL_BOX,M,N,P,m,n,p,1,1,lx,ly,lz,&da_swarm);
    DMSetFromOptions(da_swarm);
    DMSetUp(da_swarm);
    DMDAGetCorners(da_swarm,&istart,&jstart,&kstart,&isize,&jsize,&ksize);
    PetscPrintf(PETSC_COMM_SELF,"[PROC %d] istart = %d, isize = %d, jstart = %d, jsize = %d, kstart = %d, ksize= %d \n", rank, istart, isize, jstart, jsize, kstart, ksize); 

    // Set non-uniform coordinates
    PetscObjectGetComm((PetscObject)da_swarm,&comm);
    DMGetCoordinateDM(da_swarm, &cda);
    DMCreateGlobalVector(cda, &xcoor);
    VecGetArray(xcoor,&coors);

    // Use the following vector to use uniform coordinates without the Petsc function -> works!
    /*PetscReal a[11];
    a[0] = 0.;
    a[1] = 0.1; 
    a[2] = 0.2;
    a[3] = 0.3;
    a[4] = 0.4;
    a[5] = 0.5; 
    a[6] = 0.6;
    a[7] = 0.7;
    a[8] = 0.8;
    a[9] = 0.9;
    a[10] = 1.;*/

    PetscReal a[11];
    a[0] = 0.;
    a[1] = 0.05; 
    a[2] = 0.1;
    a[3] = 0.17;
    a[4] = 0.4;
    a[5] = 0.5; 
    a[6] = 0.6;
    a[7] = 0.83;
    a[8] = 0.9;
    a[9] = 0.95;
    a[10] = 1.;
    cnt = 0;
    for (int k=kstart; k< kstart + ksize; k++) {
        for (int j=jstart; j< jstart+jsize; j++) {
            for (int i=istart; i< istart + isize; i++) {
                    coors[cnt++] = a[i];
                    coors[cnt++] = a[j];
                    coors[cnt++] = a[k];
                    if (i == istart && j == jstart){
                        PetscPrintf(PETSC_COMM_SELF,"[PROC %d] z = %f \n ", rank, a[k]);
                    } // N+1 points from 0 to 1 

            }
        }
    }
    VecRestoreArray(xcoor,&coors);
    DMSetCoordinates(da_swarm,xcoor);
    PetscLogObjectParent((PetscObject)da_swarm,(PetscObject)xcoor);
    VecDestroy(&xcoor);
    DMDestroy(&cda);

    
    // OR SetUniformCoordinates:
    /*DMDASetUniformCoordinates(da_swarm,0.,1.,0.,1.,0.,1.);
    PetscReal a[11];
    a[0] = 0.;
    a[1] = 0.1; 
    a[2] = 0.2;
    a[3] = 0.3;
    a[4] = 0.4;
    a[5] = 0.5; 
    a[6] = 0.6;
    a[7] = 0.7;
    a[8] = 0.8;
    a[9] = 0.9;
    a[10] = 1.;*/

    // Create a Swarm DMDA linked to the above mesh
    DMCreate(PETSC_COMM_WORLD,&swarm);
    DMSetType(swarm,DMSWARM);
    DMSetDimension(swarm,3);
    DMSwarmSetType(swarm,DMSWARM_PIC);
    DMSwarmSetCellDM(swarm,da_swarm);
    DMSwarmFinalizeFieldRegister(swarm); 

    // Creation of a SWARM for each proc to read values
    DM swarm_init;
    PetscReal             *coords, *coords_init;
    PetscInt nbr_particle_proc = 0;
    PetscInt             *cell_id;
    PetscInt              npoints[1];

    DMCreate(PETSC_COMM_SELF, &swarm_init);
    DMSetType(swarm_init,DMSWARM);
    DMSetDimension(swarm_init,3);
    DMSwarmSetType(swarm_init,DMSWARM_PIC);
    DMSwarmSetCellDM(swarm_init,da_swarm);

	DMSwarmFinalizeFieldRegister(swarm_init);
    DMSwarmSetLocalSizes(swarm_init,part_nbr,0);
    Vec coordinates;
    DMSwarmCreateGlobalVectorFromField(swarm_init,"DMSwarmPIC_coor", &coordinates);
    char coord_path[128];
    PetscViewer coord_view;
    sprintf(coord_path,"results/-1_coord.txt");
    PetscViewerBinaryOpen(PETSC_COMM_SELF,coord_path,FILE_MODE_READ,&coord_view);
	VecLoad(coordinates,coord_view);
    DMSwarmDestroyGlobalVectorFromField(swarm_init,"DMSwarmPIC_coor",&coordinates);
    PetscViewerDestroy(&coord_view);

    /**************** Only keep the particles that are on this processor ****************/
    // 1. Minimum and maximum values:
    double x_min, x_max, y_min, y_max, z_min, z_max;
    PetscPrintf(PETSC_COMM_SELF,"[PROC %d] istart = %d, isize = %d, jstart = %d, jsize = %d, kstart = %d, ksize= %d \n", rank, istart, isize, jstart, jsize, kstart, ksize); 
    
    if(istart == 0){
        x_min = a[istart];
    }
    else{
        x_min = a[istart-1];
    }
    if(jstart == 0){
        y_min = a[jstart];
    }
    else{
        y_min = a[jstart-1];
    }
    if(kstart == 0){
        z_min = a[kstart];
    }
    else{
        z_min = a[kstart-1];
    }
    x_max = a[istart+isize-1];
    y_max = a[jstart+jsize-1];
    z_max = a[kstart+ksize-1];
    
    // 2. Number of particles on this proc:
    DMSwarmGetField(swarm_init,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords_init);
    for (int i=0; i<part_nbr ; i++){
    //PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f \n", rank, coords_init[i*3], coords_init[i*3+1], coords_init[i*3+2]);
        if (coords_init[i*3] >= x_min && coords_init[i*3] < x_max){
            if(coords_init[i*3+1] >= y_min && coords_init[i*3+1] < y_max){
                if(coords_init[i*3+2] >= z_min && coords_init[i*3+2] < z_max){
                    nbr_particle_proc++;
                }
            }
        }
    }
    DMSwarmRestoreField(swarm_init,"DMSwarmPIC_coor",NULL,NULL,(void**)&coords_init);
    PetscPrintf(PETSC_COMM_SELF,"[PROC %d] nbr_part = %d, xmin = %f, xmax = %f, ymin = %f, ymax = %f, zmin = %f, zmax = %f \n", rank, nbr_particle_proc, x_min, x_max, y_min, y_max, z_min, z_max);
    

    // 3. Assign the number of particles and a buffer of 10%
    DMSwarmSetLocalSizes(swarm, nbr_particle_proc,nbr_particle_proc*0.1);


    // 4. Copy coordinate values
    DMSwarmGetField(swarm_init,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords_init);
    DMSwarmGetField(swarm,DMSwarmPICField_coor,NULL,NULL,(void**)&coords);
    nbr_particle_proc = 0;
    for (int i=0; i<part_nbr ; i++){
        if (coords_init[i*3] >= x_min && coords_init[i*3] < x_max){
            if(coords_init[i*3+1] >= y_min && coords_init[i*3+1] < y_max){
                if(coords_init[i*3+2] >= z_min && coords_init[i*3+2] < z_max){
                    coords[nbr_particle_proc*3] = coords_init[i*3];
                    coords[nbr_particle_proc*3+1] = coords_init[i*3+1];
                    coords[nbr_particle_proc*3+2] = coords_init[i*3+2];
                    nbr_particle_proc++;
                }
            }
        }
    }
    DMSwarmRestoreField(swarm_init,"DMSwarmPIC_coor",NULL,NULL,(void**)&coords_init);
    DMSwarmRestoreField(swarm,DMSwarmPICField_coor,NULL,NULL,(void**)&coords);

    //5. Destroy the PETSC_COMM_SELF swarm
    DMDestroy(&swarm_init);

    //6. Print before Migration
    DMSwarmGetLocalSize(swarm,&npoints[0]);
    DMSwarmGetField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmGetField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);
    for (int i=0; i< npoints[0] ; i++){
		PetscPrintf(PETSC_COMM_SELF,"BEFORE [PROC %d] x = %f, y = %f, z = %f, cell_id =  %d \n", rank, coords[i*3], coords[i*3+1], coords[i*3+2], cell_id[i]);
    }
    DMSwarmRestoreField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmRestoreField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);

    //7. Migrate to have the correct cell_id
    DMSwarmMigrate(swarm,PETSC_TRUE);

    // 8: print the coordinates of each cell
    DMSwarmGetLocalSize(swarm,&npoints[0]);
    DMSwarmGetField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmGetField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);
    for (int i=0; i< npoints[0] ; i++){
		PetscPrintf(PETSC_COMM_SELF,"[PROC %d] x = %f, y = %f, z = %f, cell_id =  %d \n", rank, coords[i*3], coords[i*3+1], coords[i*3+2], cell_id[i]);
    }
    DMSwarmRestoreField(swarm,"DMSwarmPIC_coor",NULL,NULL,(void **) &coords);
    DMSwarmRestoreField(swarm,"DMSwarm_cellid",NULL,NULL,(void **) &cell_id);

    PetscInt npoints_total[1];
    DMSwarmGetSize(swarm,&npoints_total[0]);
    PetscPrintf(PETSC_COMM_SELF,"[PROC %d] nbr of cells = %d \n", rank, npoints_total[0]);
    
    DMDestroy(&da);
    //DMDestroy(&da_swarm);
    DMDestroy(&swarm);
    
}

int main(int argc,char **argv)
{
  PetscErrorCode ierr;

  ierr = PetscInitialize(&argc,&argv,(char*)0,help);if (ierr) return ierr;
  PetscPrintf(PETSC_COMM_WORLD,"------------------Give random location - migrate - write values \n");
  //ierr = ex3_1();CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"------------------Read location and field from file \n");
  //ierr = ex3_2();CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"------------------DONE! \n");
  ierr = ex3_3();CHKERRQ(ierr);
  ierr = PetscFinalize();
  return ierr;
}

/*TEST

   build:
      requires: double !complex

   test:
      filter: grep -v atomic
      filter_output: grep -v atomic
TEST*/
